# $Idaemons: /home/cvs/pkgtools/pkgdb.rb,v 1.36 2001/08/29 09:24:14 knu Exp $

require 'singleton'
require 'tsort'

class PkgDB
  include Singleton

  DB_DIR = ENV['PKG_DBDIR'] || '/var/db/pkg'

  PKGDB_FILES = {
    :contents => '+CONTENTS',
    :comment => '+COMMENT',
    :desc => '+DESC',
    :install => '+INSTALL',
    :post_install => '+POST-INSTALL',
    :deinstall => '+DEINSTALL',
    :post_deinstall => '+POST-DEINSTALL',
    :require => '+REQUIRE',
    :required_by => '+REQUIRED_BY',
    :display => '+DISPLAY',
    :mtree => '+MTREE_DIRS',
  }

  CMD = {
    :pkg_add		=> '/usr/sbin/pkg_add',
    :pkg_delete		=> '/usr/sbin/pkg_delete',
    :pkg_info		=> '/usr/sbin/pkg_info',
    :pkg_deinstall	=> '/usr/local/sbin/pkg_deinstall',
    :pkg_fetch		=> '/usr/local/sbin/pkg_fetch',
    :pkg_tarup		=> '/usr/local/sbin/pkg_tarup',
  }

  class DBError < StandardError
#    def message
#      "package database error"
#    end
  end

  attr_reader :db_dir

  def initialize(*args)
    setup(*args)
  end

  def setup(db_dir = nil)
    @db_dir = File.expand_path(db_dir || DB_DIR)
    @abs_db_dir = nil
    self
  end

  def abs_db_dir
    unless @abs_db_dir
      pwd = Dir.pwd
      Dir.chdir(@db_dir) rescue raise DBError.new("Can't chdir to '#{@db_dir}'")

      @abs_db_dir = Dir.pwd

      Dir.chdir(pwd)
    end

    @abs_db_dir
  end

  def strip(path, installed_only = false)
    base = path.chomp('/')	# allow `pkgname/'

    if base.include?('/')
      dir, base = File.split(File.expand_path(base))

      if dir != @db_dir && dir != abs_db_dir
	return nil
      end
    end

    if installed_only && !installed?(base)
      return nil
    end

    base
  end

  def pkgdir(pkgname)
    File.join(@db_dir, pkgname)
  end

  def pkgdir?(dir)
    File.exist?(File.join(dir, PKGDB_FILES[:contents]))
  end

  def pkgfile(pkgname, filename)
    if filename.kind_of?(Symbol)
      filename = PKGDB_FILES[filename]
    end

    File.join(pkgdir(pkgname), filename)
  end

  def pkg(pkgname)
    installed?(pkgname) or return nil

    PkgInfo.new(pkgname) rescue nil
  end

  def origin_of(pkgname)
    p = pkg(pkgname)

    if p
      p.origin
    else
      nil
    end
  end

  def installed?(pkgname)
    pkgdir?(pkgdir(pkgname))
  end

  def required?(pkgname)
    File.exist?(pkg_required_by(pkgname))
  end

  def required_by(pkgname)
    filename = pkg_required_by(pkgname)

    File.exist?(filename) or return nil

    return File.open(filename).readlines.collect {
      |line| line.chomp
    }
  end

  def pkgdep(pkgname)
    filename = pkg_contents(pkgname)

    File.exist?(filename) or return nil

    dep = []

    File.open(filename).each_line { |line|
      if /^@pkgdep +(.*)/ =~ line
	dep << $1
      end
    }

    dep
  end

  PKGDB_FILES.each_key do |key|
    module_eval %{
      def pkg_#{key.to_s}(pkgname)
	pkgfile(pkgname, #{key.inspect})
      end
    }
  end

  def glob(pattern = '*', want_pkg_info = true)
    list = []
    pkg = nil

    # shortcut
    if installed?(pattern)
      pkgname = pattern

      begin
	pkg = PkgInfo.new(pkgname)
      rescue => e
	STDERR.puts e.message + ": skipping..."
	return list
      end

      if block_given?
	yield(want_pkg_info ? pkg : pkgname)
	return nil
      else
	return [want_pkg_info ? pkg : pkgname]
      end
    end      

    Dir.entries(@db_dir).sort.each do |pkgname|
      installed?(pkgname) or next

      begin
	pkg = PkgInfo.new(pkgname)
      rescue => e
	STDERR.puts e.message + ": skipping..."
	next
      end

      pkg.match?(pattern) or next

      if block_given?
	yield(want_pkg_info ? pkg : pkgname)
      else
	list.push(want_pkg_info ? pkg : pkgname)
      end
    end rescue raise DBError.new("Can't open '#{@db_dir}'")

    if block_given?
      nil
    else
      list
    end
  end

  def tsort(pkgnames)
    t = TSort.new

    pkgnames.each do |pkgname|
      deps = pkgdep(pkgname) || []

      t.add(pkgname, *deps)
    end

    t
  end

  def sort(pkgnames)
    tsort(pkgnames).tsort!.select { |pkgname|
      pkgnames.include?(pkgname)
    }
  end

  def sort!(pkgnames)
    pkgnames.replace(sort(pkgnames))
  end

  def recurse(pkgname, recurse_down = false, recurse_up = false)
    list = []

    if recurse_up
      deps = pkgdep(pkgname)

      if deps
	deps.each { |name| list << name if installed?(name) }
      end
    end

    list << pkgname

    if recurse_down
      deps = required_by(pkgname)

      if deps
	deps.each { |name| list << name if installed?(name) }
      end
    end

    list
  end
end
