# $Idaemons: /home/cvs/pkgtools/portsdb.rb,v 1.38 2001/08/31 07:44:09 knu Exp $

require 'singleton'
require 'tempfile'
require 'dbm' unless defined?(DBM)
require 'fnmatch' unless File.respond_to?(:fnmatch?)

class PortsDB
  include Singleton
  include Enumerable

  DB_VERSION = [:FreeBSD, 2]

  PORTS_DIR = ENV['PORTSDIR'] || '/usr/ports'
  PORTS_DBDIR = ENV['PORTS_DBDIR'] || PORTS_DIR
  PORTS_DBDIR_LIST = [
    ENV['PORTS_DBDIR'],
    PORTS_DIR,
    PkgDB::DB_DIR,
    ENV['TMPDIR'],
    '/var/tmp',
    '/tmp'
  ].compact

  MY_PORT = 'sysutils/portupgrade'

  attr_reader :ports_dir, :index_file, :db_dir, :db_file

  class IndexFileError < StandardError
#    def message
#      "index file error"
#    end
  end

  class DBError < StandardError
#    def message
#      "database file error"
#    end
  end

  def setup(db_dir = nil, ports_dir = nil)
    @ports_dir = ports_dir || PORTS_DIR
    @abs_ports_dir = nil
    @index_file = ENV['PORTS_INDEX'] || File.join(@ports_dir, 'INDEX')
    @db = nil

    set_db_dir(db_dir || PORTS_DBDIR)

    @categories = nil
    @virtual_categories = nil
    @origins = nil
    @pkgnames = nil
    @origins_by_categories = {}
    @ports = {}
    @localbase = nil
    @x11base = nil
    self
  end

  def set_db_dir(db_dir)
    @db_dir = db_dir

    @db_file = File.join(db_dir, 'INDEX.db')
    @db_filebase = @db_file.sub(/\.db$/, '')
    @rbo_file = File.join(db_dir, 'INDEX.rbo')

    close_db
  end
  private :set_db_dir

  def abs_ports_dir
    if @abs_ports_dir.nil?
      pwd = Dir.pwd
      Dir.chdir(@ports_dir)

      @abs_ports_dir = Dir.pwd

      Dir.chdir(pwd)
    end

    @abs_ports_dir
  end

  def localbase
    @localbase ||= `cd #{@ports_dir}/#{MY_PORT} && make -V LOCALBASE`.strip
  end

  def x11base
    @x11base ||= `cd #{@ports_dir}/#{MY_PORT} && make -V X11BASE`.strip
  end

  def join(category, port)
    File.join(category, port)
  end

  def split(origin)
    if %r"^([^./A-Z][^/]*)/([^./][^/]*)$" =~ path
      return $1, $2
    end

    nil
  end

  def strip(path, existing_only = false)
    path = path.chomp('/')	# allow `category/ports/'

    if %r"^[^./][^/]*/[^./][^/]*$" =~ path
      if existing_only && !exists?(path)
	return nil
      end

      return path
    end

    parentdir, portname = File.split(File.expand_path(path))

    parentdir == '/' and return nil

    dir, category = File.split(parentdir)

    port = join(category, portname)

    if existing_only
      exists?(port) and return port
    else
      (dir == @ports_dir || dir == abs_ports_dir) and return port
    end

    nil
  end

  def portdir(port)
    File.join($ports_dir, port)
  end

  def categories
    open_db if @categories.nil?

    @categories
  end

  def virtual_categories
    open_db if @virtual_categories.nil?

    @virtual_categories
  end

  def update(make_args = "-j4")
    STDERR.print "Updating the ports index ..."
    STDERR.flush

    curdir = `cd #{@ports_dir} && make -V .CURDIR`.chomp

    t = Tempfile.new('INDEX')
    t.close
    tmp = t.path

    system %[
      cd #{curdir} &&
      make #{make_args} describe ECHO_MSG="echo > /dev/null" |
      perl #{curdir}/Tools/make_index |
      sed -e 's./.\001.g' |
      sort -t '|' +1 -2 |
      sed -e 's.\001./.g' > #{tmp}
    ]
      
    if File.zero?(tmp)
      STDERR.puts 'failed to generate INDEX!'
      raise IndexFileError, "index generation error"
    end

    begin
      File.chmod(0644, tmp)
    rescue => e
      STDERR.puts e.message
      raise IndexFileError, "index chmod error"
    end

    if not system('/bin/mv', '-f', tmp, @index_file)
      STDERR.puts 'failed to overwrite INDEX!"'
      raise IndexFileError, "index overwrite error"
    end

    STDERR.puts " done"

    @categories = nil
    @virtual_categories = nil
    @origins = nil
    @pkgnames = nil
    @origins_by_categories = {}
    @ports = {}

    close_db
  end

  def open_db
    @db and return @db

    update_db

    retried = false

    begin
      File.open(@rbo_file) do |f|
	db_version = Marshal.load(f)

	if db_version[0] != DB_VERSION[0] || db_version[1] < DB_VERSION[1]
	  f.close
	  raise TypeError
	end

	@categories = Marshal.load(f)
	@virtual_categories = Marshal.load(f)
	@origins = Marshal.load(f)
	@pkgnames = Marshal.load(f)
	@origins_by_categories = Marshal.load(f)
      end
    rescue TypeError => e
      update_db(true)

      if retried
	raise DBError, "#{e.message}: Cannot read the ports DB! (#{@rbo_file})"
      else
	retried = true
	retry
      end
    end

    begin
      @db = DBM.open(@db_filebase)
    rescue => e
      raise DBError, "#{e.message}: Cannot open the ports DB! (#{@db_file})"
    end

    @db
  rescue => e
    STDERR.puts e.message
    raise DBError, 'database file error'
  end

  def close_db
    if @db
      @db.close
      @db = nil
    end
  end

  def up_to_date?
    stat_index = File.stat(@index_file)

    return File.exist?(@db_file) && File.stat(@db_file) >= stat_index &&
      File.exist?(@rbo_file) && File.stat(@rbo_file) >= stat_index
  end

  def update_db(force = false)
    update if not File.exists?(@index_file)

    if force
      close_db
    else
      up_to_date? and return false

      close_db

      File.writable?(@db_dir) or PORTS_DBDIR_LIST.each do |dir|
	set_db_dir(dir)

	up_to_date? and return false

	File.writable?(dir) and break
      end
    end

    prev_sync = STDERR.sync
    STDERR.sync = true

    STDERR.print "Updating the ports database in #{@db_dir} "

    @categories = `cd #{@ports_dir} && make -V SUBDIR`.split

    i = -1

    @origins = []
    @pkgnames = []

    begin
      File.unlink(@db_file) if File.exist?(@db_file)

      @db = DBM.open(@db_filebase, 0664)

      File.open(@index_file) do |f|
	f.each_with_index do |line, i|
	  lineno = i + 1

	  STDERR.print '.' if lineno % 500 == 0

	  begin
	    port_info = PortInfo.new(line)
	    origin = port_info.origin
	    pkgname = port_info.pkgname

	    port_info.categories.each do |category|
	      if @origins_by_categories.key?(category)
		@origins_by_categories[category] << origin
	      else
		@origins_by_categories[category] = [origin]
	      end
	    end

	    @origins << origin
	    @pkgnames << pkgname

	    @db[origin] = port_info
	    @db[pkgname.to_s] = port_info
	  rescue => e
	    STDERR.puts @index_file + ":#{lineno}:#{e.message}"
	  end
	end
      end
    rescue => e
      File.unlink(@db_file) if File.exist?(@db_file)
      raise DBError, "#{e.message}: Cannot update the ports DB! (#{@db_file})"
    ensure
      close_db
    end

    @virtual_categories = (@origins_by_categories.keys - @categories).sort!

    begin
      File.unlink(@rbo_file) if File.exist?(@rbo_file)

      File.open(@rbo_file, 'w') do |f|
	Marshal.dump(DB_VERSION, f)
	Marshal.dump(@categories, f)
	Marshal.dump(@virtual_categories, f)
	Marshal.dump(@origins, f)
	Marshal.dump(@pkgnames, f)
	Marshal.dump(@origins_by_categories, f)
      end
    rescue => e
      File.unlink(@rbo_file) if File.exist?(@rbo_file)
      raise DBError, "#{e.message}: Cannot update the ports DB! (#{@rbo_file})"
    end

    STDERR.puts "done"
    STDOUT.sync = prev_sync

    true
  end

  def port(key)
    key.is_a?(PortInfo) and return key

    @ports[key] and return @ports[key]

    open_db

    line = @db[key]

    @ports[key] = if line then PortInfo.new(line) else nil end
  end

  alias [] port

  def origin_of(key)
    p = port(key)

    if p
      p.origin
    else
      nil
    end
  end

  def origins(category = nil)
    open_db

    if category
      @origins_by_categories[category]
    else
      @origins
    end
  end

  def each(category = nil)
    origins(category).each { |key|
      yield(@db[key])
    }
  end

  def each_origin(category = nil)
    origins(category).each { |key|
      yield(key)
    }
  end

  def each_pkgname
    open_db

    @pkgnames.each { |key|
      yield(key)
    }
  end

  def glob(pattern = '*')
    list = []
    pkg = nil

    open_db

    case pattern
    when Regexp
      is_port = pattern.source.include?('/')
    else
      is_port = pattern.include?('/')

      # shortcut
      portinfo = port(pattern)

      if portinfo
	if block_given?
	  yield(portinfo)
	  return nil
	else
	  return [portinfo]
	end
      end
    end

    if is_port
      @origins.each do |origin|
	case pattern
	when Regexp
	  next if pattern !~ origin
	else
	  next if not File.fnmatch?(pattern, origin, File::FNM_PATHNAME)
	end

	portinfo = port(origin)

	if portinfo
	  if block_given?
	    yield(portinfo)
	  else
	    list.push(portinfo)
	  end
	end
      end
    else
      @pkgnames.each do |pkgname|
	next if not pkgname.match?(pattern)

	portinfo = port(pkgname.to_s)

	if portinfo
	  if block_given?
	    yield(portinfo)
	  else
	    list.push(portinfo)
	  end
	end
      end
    end

    if block_given?
      nil
    else
      list
    end
  end

  def exists?(port)
    return if %r"^[^/]+/[^/]+$" !~ port

    dir = portdir(port)

    return false if not File.directory?(dir) &&
      File.file?(File.join(dir, 'Makefile'))

    pkgname = `cd #{dir} && make -V PKGNAME 2>/dev/null`.chomp

    if pkgname.empty?
      false
    else
      pkgname
    end
  end

  def masters(port)
    dir = portdir(port)

    ports = []

    `cd #{dir} ; make -dd -n 2>&1`.each do |line|
      if /^Searching for .*\.\.\.Caching .* for (\S+)/ =~ line.chomp
	path = File.expand_path($1)

	if (path.sub!(%r"^#{Regexp.quote(@ports_dir)}/", '') ||
	    path.sub!(%r"^#{Regexp.quote(abs_ports_dir)}/", '')) &&
	    exists?(path)
	  ports << path unless ports.include?(path)
	end
      end
    end

    ports.delete(port)

    ports
  end
  
  def sort(ports)
    tsort = TSort.new

    ports.each do |p|
      portinfo = port(p)

      portinfo or next

      deps = portinfo.all_depends().collect { |i| origin_of(i) } || []

      tsort.add(portinfo.origin, *deps)
    end

    tsort.tsort!.select { |p|
      ports.include?(p)
    }
  end

  def sort!(ports)
    ports.replace(sort(ports))
  end

  def recurse(portinfo, recurse_down = false, recurse_up = false)
    if not portinfo.is_a?(PortInfo)
      portinfo = port(portinfo)
    end

    list = []

    portinfo or return list

    if recurse_up
      portinfo.all_depends.map do |name|
	i = port(name)

	list << i if i
      end
    end

    list << portinfo

    if recurse_down
      # slow!
      pkgname = portinfo.pkgname.fullname

      glob do |i|
	list << i if i.all_depends.include?(pkgname)
      end
    end

    list
  end
end
